#include <pic.h>
#include <string.h>
#include "firmware.h"

#define Set_PIC_Ack(/* bit */ val) (val ?  PORTC |= 0x10 : PORTC &= 0xef)
#define /* bit */ Get_PC_Req() (PORTC & 0x01)
#define Wait_For_PC_Req(/* bit */ val) { while ( Get_PC_Req() != val ) ; }

#define READ_SHORT(/* short */ dest) \
  Wait_For_PC_Req(0); /* lsb */ \
  dest = PORTB; \
  Set_PIC_Ack(0); /* read complete */ \
  Wait_For_PC_Req(1); /* msb */ \
  dest |= (PORTB<<8); \
  Set_PIC_Ack(1) /* read complete */

#define WRITE_SHORT(/* short */ source) \
  PORTB = source&0xff; \
  Set_PIC_Ack(0); /* data stable */ \
  Wait_For_PC_Req(1); /* read complete */ \
  PORTB = (source>>8)&0xff; \
  Set_PIC_Ack(1); /* data stable */ \
  Wait_For_PC_Req(0) /* read complete */

const static char *Version_Info = VERSION_INFO;

struct_glob glob;
unsigned char k_index;

static void Init_Globals(void) {
  glob.stamp = 0;
  glob.step_1 = glob.step_2 = 1;
  glob.attack = 1;
  glob.sustain = 1;
  glob.pause = 1;
  glob.shift = 2;
  glob.shift_mode = SHIFT_DN_MID;
  glob.state = ENV_OFF;
  glob.on_mode = ON_MODE_SUSTAIN;
  glob.freq = 255;
}

static void Process_OP_On(void) {
  Set_PIC_Ack(1); /* acknowledges opcode */
  glob.state = ENV_INIT;
  Wait_For_PC_Req(0); Set_PIC_Ack(0);
}

static void Process_OP_Off(void) {
  Set_PIC_Ack(1);
  glob.state = ENV_OFF;
  Wait_For_PC_Req(0); Set_PIC_Ack(0);
}

static void Process_OP_Get_State(void) {
  Set_PIC_Ack(1); /* acknowledges opcode */
  Wait_For_PC_Req(0); /* parallel port in input mode */
  TRISB = 0x00; /* PORTB output */
  PORTB = glob.state;
  Set_PIC_Ack(0);
  Wait_For_PC_Req(1); /* has read it */
  TRISB = 0xff; /* PORTB input */
  Set_PIC_Ack(1);
  Wait_For_PC_Req(0); /* parallel port in output mode */
  Set_PIC_Ack(0);
}

static void Process_OP_Inc(void) {
  char arg;
  Set_PIC_Ack(1); /* acknowledges opcode */
  Wait_For_PC_Req(0); arg = PORTB; Set_PIC_Ack(0);
  Wait_For_PC_Req(1); /* wait for port to become input */
  TRISB = 0x00; /* turn PORTB into output */
  PORTB = arg+1;
  Set_PIC_Ack(1); /* signals write is complete */
  Wait_For_PC_Req(0); /* wait till arg is read */
  TRISB = 0xff; /* turn PORTB into input */
  Set_PIC_Ack(0);
}

static void Send_Stamp_And_Four_Samples(void) {
  TRISB = 0x00; /* turn PORTB into output */
  PORTB = glob.stamp&0xff; 
  Set_PIC_Ack(0); Wait_For_PC_Req(1); /* PC has read it */
  PORTB = glob.stamp>>8;
  Set_PIC_Ack(1); Wait_For_PC_Req(0); /* PC has read it */
  TRISB = 0xff; /* now A/D talks ! */

  PORTA = 0b001000; /* ~CS -> 0 */

  PORTA = 0b001000; /* ~RD -> 1, A[1:0] = 00 */
  PORTA = 0b000000; /* ~RD -> 0 */
  Set_PIC_Ack(0); /* sample_1 is stable */
  Wait_For_PC_Req(1); /* PC has read it */

  PORTA = 0b001010; /* ~RD -> 1, A[1:0] = 01 */
  PORTA = 0b000010; /* ~RD -> 0 */
  Set_PIC_Ack(1); /* sample_2 is stable */
  Wait_For_PC_Req(0); /* PC has read it */

  PORTA = 0b001100; /* ~RD -> 1, A[1:0] = 10 */
  PORTA = 0b000100; /* ~RD -> 0 */
  Set_PIC_Ack(0); /* sample_3 is stable */
  Wait_For_PC_Req(1); /* PC has read it */

  PORTA = 0b001110; /* ~RD -> 1, A[1:0] = 11 */
  PORTA = 0b000110; /* ~RD -> 0 */
  Set_PIC_Ack(1); /* sample_4 is stable */
  Wait_For_PC_Req(0); /* PC has read it */

  PORTA = 0b101000; /* ~CS -> 1 */
}

static void Process_OP_Sample(void) {
  Set_PIC_Ack(1); /* acknowledges opcode */
  Wait_For_PC_Req(0); /* waits for PC to turn port around */
  Send_Stamp_And_Four_Samples();
  Set_PIC_Ack(0);
}

static void Process_OP_Sample_Batch(unsigned char sync) {
  unsigned short repeat, last_stamp;
  Set_PIC_Ack(1); Wait_For_PC_Req(0);
  repeat = PORTB;
  Set_PIC_Ack(0); Wait_For_PC_Req(1);
  repeat |= (PORTB<<8);
  Set_PIC_Ack(1); Wait_For_PC_Req(0);

  if (sync && (glob.state != ENV_OFF)) /* k is incremented? */
    while (k_index) ; /* wait for beginning of cycle */
  glob.stamp = last_stamp = 0;
  while (repeat--) {
    while (glob.stamp == last_stamp) ;
    last_stamp = glob.stamp;
    Send_Stamp_And_Four_Samples();
  }

  Set_PIC_Ack(0);
}

static void Process_OP_Set_Envelope(void) {
  Set_PIC_Ack(1); /* acknowledges opcode */

  READ_SHORT(glob.attack);
  READ_SHORT(glob.sustain);
  READ_SHORT(glob.pause);
  Wait_For_PC_Req(0); glob.shift      = PORTB; Set_PIC_Ack(0);
  Wait_For_PC_Req(1); glob.shift_mode = PORTB; Set_PIC_Ack(1);
  Wait_For_PC_Req(0); Set_PIC_Ack(0); /* final handshake */
}

static void Process_OP_Set_Steps(void) {
  Set_PIC_Ack(1); /* acknowledges opcode */
  READ_SHORT(glob.step_1);
  READ_SHORT(glob.step_2);
  Wait_For_PC_Req(0); Set_PIC_Ack(0);
}

static void Process_OP_Set_On_Mode(void) {
  Set_PIC_Ack(1); /* acknowledges opcode */
  Wait_For_PC_Req(0); glob.on_mode = PORTB; Set_PIC_Ack(0);
}

static void Process_OP_Set_Freq(void) {
  Set_PIC_Ack(1); /* acknowledges opcode */
  Wait_For_PC_Req(0); glob.freq = PORTB; Set_PIC_Ack(0);
}

static void Send_Bytes(unsigned char n, const char *byte_ptr) {
  char i, ack, req;
  Set_PIC_Ack(1); /* acknowledges opcode */

  Wait_For_PC_Req(0); /* data is input */
  TRISB = 0x00;

  PORTB = n;
  Set_PIC_Ack(0); /* data stable */
  Wait_For_PC_Req(1); /* read complete */

  for (i = 0, ack = 1, req = 0; i < n;
       i++, ack = !ack, req = !req, byte_ptr++) {
    PORTB = *byte_ptr;
    Set_PIC_Ack(ack); /* data stable */
    Wait_For_PC_Req(req); /* read complete */
  }
  TRISB = 0xff;

  if (!(n & 0x01)) {/* even */
    Set_PIC_Ack(1);
    Wait_For_PC_Req(0);
  }
  Set_PIC_Ack(0);
}

static void Process_OP_Get_Glob(void) {
  Send_Bytes(sizeof(struct_glob), (const char *)&glob);
}

static void Process_OP_Get_Version_Info(void) {
  Send_Bytes(strlen(Version_Info)+1, (const char *)Version_Info);
}

bank2 signed char glob_table_T[CYCLE_SAMPLES];
bank3 signed char glob_table_R[CYCLE_SAMPLES];

static void Process_OP_Set_Wave_T(void) {
  char i, req;
  Set_PIC_Ack(1); /* acknowledges opcode */
  for (i = 0, req = 0; i < CYCLE_SAMPLES; i++, req = !req) {
    Wait_For_PC_Req(req); glob_table_T[i] = PORTB; Set_PIC_Ack(req);
  }
  Wait_For_PC_Req(0); Set_PIC_Ack(0); /* assumes CYCLE_SAMPLES is even! */
}

static void Process_OP_Set_Wave_R(void) {
  char i, req;
  Set_PIC_Ack(1); /* acknowledges opcode */
  for (i = 0, req = 0; i < CYCLE_SAMPLES; i++, req = !req) {
    Wait_For_PC_Req(req); glob_table_R[i] = PORTB; Set_PIC_Ack(req);
  }
  Wait_For_PC_Req(0); Set_PIC_Ack(0); /* assumes CYCLE_SAMPLES is even! */
}

static void Process_OP_Get_Wave_T(void) {
  Send_Bytes(CYCLE_SAMPLES, (const char *)&glob_table_T);
}

static void Process_OP_Get_Wave_R(void) {
  Send_Bytes(CYCLE_SAMPLES, (const char *)&glob_table_R);
}

static void Init_PIC(void) {
  TRISA = 0x01; /* A/D control: RA{5:0} = {~CS,CONVST,~RD,A1,A0,~BUSY} */ 
  ADCON1 = 0x07;
  PORTA = 0b101000;

  TRISC = 0x07; /* RC4 = PIC_Ack; RC7,6,5,3 = debug */

  INTCON = 0x00;   // disable all the interrupts
  PIE1 = 0x00;     // disables all other interrupts
  PR2 = 0xFF;      // sets period of TMR2 
  CCPR1L = 0x3F;   // sets MSB of the duty cycle
  CCPR2L = 0x3F;   // sets MSB of the duty cycle
  TRISC &= 0xf9;   // Make CCP2 (RC1) & CCP1 (RC2) outputs
  // T2CON = 0x08; // Prescale 1:1 and postscale is 1:2
  T2CON = 0x01; // Prescale 1:4 and postscale is 1:1
  CCP1CON = 0x0F;  // PWM mode; set LSB of duty cycle 
  CCP2CON = 0x0F;  // PWM mode; set LSB of duty cycle 
  CCP1IE = 0;
  CCP2IE = 0;
  TMR2ON = 1;      // turn the timer 2 on
  TMR2IE = 1;      // enables the TMR2 intr
  PEIE = 1;
  GIE = 1;
}

void main(void) {
  char opcode;
  Init_Globals();
  Init_PIC();
  Set_PIC_Ack(0);
  Wait_For_PC_Req(0);
  while(1) {
    Wait_For_PC_Req(1); /* opcode on bus */
    opcode = PORTB;  /* reads opcode */
    switch (opcode) {
      case OP_On:               Process_OP_On();               break;
      case OP_Off:              Process_OP_Off();              break;
      case OP_Inc:              Process_OP_Inc();              break;
      case OP_Sample:           Process_OP_Sample();           break;
      case OP_Sample_Batch:     Process_OP_Sample_Batch(0);    break;
      case OP_Sample_Batch_Sync:Process_OP_Sample_Batch(1);    break;
      case OP_Set_Wave_T:       Process_OP_Set_Wave_T();       break;
      case OP_Set_Wave_R:       Process_OP_Set_Wave_R();       break;
      case OP_Get_State:        Process_OP_Get_State();        break;
      case OP_Set_Envelope:     Process_OP_Set_Envelope();     break;
      case OP_Set_Steps:        Process_OP_Set_Steps();        break;
      case OP_Set_On_Mode:      Process_OP_Set_On_Mode();      break;
      case OP_Set_Freq:         Process_OP_Set_Freq();         break;
      case OP_Get_Glob:         Process_OP_Get_Glob();         break;
      case OP_Get_Wave_T:       Process_OP_Get_Wave_T();       break;
      case OP_Get_Wave_R:       Process_OP_Get_Wave_R();       break;
      case OP_Get_Version_Info: Process_OP_Get_Version_Info(); break;
    }
  }
}
